@@ -1,7 +1,7 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Nils Weiss <nils@we155.de>
-# This program is published under a GPLv2 license
 
 # scapy.contrib.description = Unified Diagnostic Service (UDS)
 # scapy.contrib.status = loads
@@ -9,15 +9,15 @@
 import struct
 import time
 
-from itertools import product
+from scapy.contrib.automotive import log_automotive
 from scapy.fields import ByteEnumField, StrField, ConditionalField, \
     BitEnumField, BitField, XByteField, FieldListField, \
     XShortField, X3BytesField, XIntField, ByteField, \
     ShortField, ObservableDict, XShortEnumField, XByteEnumField, StrLenField, \
-    FieldLenField
+    FieldLenField, XStrFixedLenField, XStrLenField
 from scapy.packet import Packet, bind_layers, NoPayload
 from scapy.config import conf
-from scapy.error import log_loading
+from scapy.error import log_loading, Scapy_Exception
 from scapy.utils import PeriodicSenderThread
 from scapy.contrib.isotp import ISOTP
 from scapy.compat import Dict, Union
@@ -49,6 +49,7 @@ class UDS(ISOTP):
          0x24: 'ReadScalingDataByIdentifier',
          0x27: 'SecurityAccess',
          0x28: 'CommunicationControl',
+         0x29: 'Authentication',
          0x2A: 'ReadDataPeriodicIdentifier',
          0x2C: 'DynamicallyDefineDataIdentifier',
          0x2E: 'WriteDataByIdentifier',
@@ -70,6 +71,7 @@ class UDS(ISOTP):
          0x64: 'ReadScalingDataByIdentifierPositiveResponse',
          0x67: 'SecurityAccessPositiveResponse',
          0x68: 'CommunicationControlPositiveResponse',
+         0x69: 'AuthenticationPositiveResponse',
          0x6A: 'ReadDataPeriodicIdentifierPositiveResponse',
          0x6C: 'DynamicallyDefineDataIdentifierPositiveResponse',
          0x6E: 'WriteDataByIdentifierPositiveResponse',
@@ -182,7 +184,7 @@ class UDS_ERPR(Packet):
     ]
 
     def answers(self, other):
-        return isinstance(other, UDS_ER)
+        return isinstance(other, UDS_ER) and other.resetType == self.resetType
 
 
 bind_layers(UDS, UDS_ERPR, service=0x51)
@@ -274,6 +276,162 @@ class UDS_CCPR(Packet):
 bind_layers(UDS, UDS_CCPR, service=0x68)
 
 
+# #########################AUTH###################################
+class UDS_AUTH(Packet):
+    subFunctions = {
+        0x00: 'deAuthenticate',
+        0x01: 'verifyCertificateUnidirectional',
+        0x02: 'verifyCertificateBidirectional',
+        0x03: 'proofOfOwnership',
+        0x04: 'transmitCertificate',
+        0x05: 'requestChallengeForAuthentication',
+        0x06: 'verifyProofOfOwnershipUnidirectional',
+        0x07: 'verifyProofOfOwnershipBidirectional',
+        0x08: 'authenticationConfiguration',
+        0x7F: 'ISOSAEReserved'
+    }
+    name = "Authentication"
+    fields_desc = [
+        ByteEnumField('subFunction', 0, subFunctions),
+        ConditionalField(XByteField('communicationConfiguration', 0),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02, 0x5]),
+        ConditionalField(XShortField('certificateEvaluationId', 0),
+                         lambda pkt: pkt.subFunction == 0x04),
+        ConditionalField(XStrFixedLenField('algorithmIndicator', 0, length=16),
+                         lambda pkt: pkt.subFunction in [0x05, 0x06, 0x07]),
+        ConditionalField(FieldLenField('lengthOfCertificateClient', None,
+                                       fmt="H", length_of='certificateClient'),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02]),
+        ConditionalField(XStrLenField('certificateClient', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfCertificateClient),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02]),
+        ConditionalField(FieldLenField('lengthOfProofOfOwnershipClient', None,
+                                       fmt="H",
+                                       length_of='proofOfOwnershipClient'),
+                         lambda pkt: pkt.subFunction in [0x03, 0x06, 0x07]),
+        ConditionalField(XStrLenField('proofOfOwnershipClient', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfProofOfOwnershipClient),
+                         lambda pkt: pkt.subFunction in [0x03, 0x06, 0x07]),
+        ConditionalField(FieldLenField('lengthOfChallengeClient', None,
+                                       fmt="H", length_of='challengeClient'),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02, 0x06,
+                                                         0x07]),
+        ConditionalField(XStrLenField('challengeClient', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfChallengeClient),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02, 0x06,
+                                                         0x07]),
+        ConditionalField(FieldLenField('lengthOfEphemeralPublicKeyClient',
+                                       None, fmt="H",
+                                       length_of='ephemeralPublicKeyClient'),
+                         lambda pkt: pkt.subFunction == 0x03),
+        ConditionalField(XStrLenField('ephemeralPublicKeyClient', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfEphemeralPublicKeyClient),
+                         lambda pkt: pkt.subFunction == 0x03),
+        ConditionalField(FieldLenField('lengthOfCertificateData', None,
+                                       fmt="H", length_of='certificateData'),
+                         lambda pkt: pkt.subFunction == 0x04),
+        ConditionalField(XStrLenField('certificateData', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfCertificateData),
+                         lambda pkt: pkt.subFunction == 0x04),
+        ConditionalField(FieldLenField('lengthOfAdditionalParameter', None,
+                                       fmt="H",
+                                       length_of='additionalParameter'),
+                         lambda pkt: pkt.subFunction in [0x06, 0x07]),
+        ConditionalField(XStrLenField('additionalParameter', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfAdditionalParameter),
+                         lambda pkt: pkt.subFunction in [0x06, 0x07]),
+    ]
+
+
+bind_layers(UDS, UDS_AUTH, service=0x29)
+
+
+class UDS_AUTHPR(Packet):
+    authenticationReturnParameterTypes = {
+        0x00: 'requestAccepted',
+        0x01: 'generalReject',
+        # Authentication with PKI Certificate Exchange (ACPE)
+        0x02: 'authenticationConfigurationAPCE',
+        # Authentication with Challenge-Response (ACR)
+        0x03: 'authenticationConfigurationACRWithAsymmetricCryptography',
+        0x04: 'authenticationConfigurationACRWithSymmetricCryptography',
+        0x05: 'ISOSAEReserved',
+        0x0F: 'ISOSAEReserved',
+        0x10: 'deAuthenticationSuccessful',
+        0x11: 'certificateVerifiedOwnershipVerificationNecessary',
+        0x12: 'ownershipVerifiedAuthenticationComplete',
+        0x13: 'certificateVerified',
+        0x14: 'ISOSAEReserved',
+        0x9F: 'ISOSAEReserved',
+        0xFF: 'ISOSAEReserved'
+    }
+    name = 'AuthenticationPositiveResponse'
+    fields_desc = [
+        ByteEnumField('subFunction', 0, UDS_AUTH.subFunctions),
+        ByteEnumField('returnValue', 0, authenticationReturnParameterTypes),
+        ConditionalField(XStrFixedLenField('algorithmIndicator', 0, length=16),
+                         lambda pkt: pkt.subFunction in [0x05, 0x06, 0x07]),
+        ConditionalField(FieldLenField('lengthOfChallengeServer', None,
+                                       fmt="H", length_of='challengeServer'),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02, 0x05]),
+        ConditionalField(XStrLenField('challengeServer', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfChallengeServer),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02, 0x05]),
+        ConditionalField(FieldLenField('lengthOfCertificateServer', None,
+                                       fmt="H", length_of='certificateServer'),
+                         lambda pkt: pkt.subFunction == 0x02),
+        ConditionalField(XStrLenField('certificateServer', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfCertificateServer),
+                         lambda pkt: pkt.subFunction == 0x02),
+        ConditionalField(FieldLenField('lengthOfProofOfOwnershipServer', None,
+                                       fmt="H",
+                                       length_of='proofOfOwnershipServer'),
+                         lambda pkt: pkt.subFunction in [0x02, 0x07]),
+        ConditionalField(XStrLenField('proofOfOwnershipServer', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfProofOfOwnershipServer),
+                         lambda pkt: pkt.subFunction in [0x02, 0x07]),
+        ConditionalField(FieldLenField('lengthOfSessionKeyInfo', None, fmt="H",
+                                       length_of='sessionKeyInfo'),
+                         lambda pkt: pkt.subFunction in [0x03, 0x06, 0x07]),
+        ConditionalField(XStrLenField('sessionKeyInfo', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfSessionKeyInfo),
+                         lambda pkt: pkt.subFunction in [0x03, 0x06, 0x07]),
+        ConditionalField(FieldLenField('lengthOfEphemeralPublicKeyServer',
+                                       None, fmt="H",
+                                       length_of='ephemeralPublicKeyServer'),
+                         lambda pkt: pkt.subFunction in [0x01, 0x02]),
+        ConditionalField(XStrLenField('ephemeralPublicKeyServer', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfEphemeralPublicKeyServer),
+                         lambda pkt: pkt.subFunction in [0x1, 0x02]),
+        ConditionalField(FieldLenField('lengthOfNeededAdditionalParameter',
+                                       None, fmt="H",
+                                       length_of='neededAdditionalParameter'),
+                         lambda pkt: pkt.subFunction == 0x05),
+        ConditionalField(XStrLenField('neededAdditionalParameter', b"",
+                                      length_from=lambda p:
+                                      p.lengthOfNeededAdditionalParameter),
+                         lambda pkt: pkt.subFunction == 0x05),
+    ]
+
+    def answers(self, other):
+        return isinstance(other, UDS_AUTH) \
+            and other.subFunction == self.subFunction
+
+
+bind_layers(UDS, UDS_AUTHPR, service=0x69)
+
+
 # #########################TP###################################
 class UDS_TP(Packet):
     name = 'TesterPresent'
@@ -338,10 +496,23 @@ bind_layers(UDS, UDS_ATPPR, service=0xC3)
 
 
 # #########################SDT###################################
+# TODO: Implement correct internal message service handling here,
+# instead of using just the dataRecord
 class UDS_SDT(Packet):
     name = 'SecuredDataTransmission'
     fields_desc = [
-        StrField('securityDataRequestRecord', b"")
+        BitField('requestMessage', 0, 1),
+        BitField('ISOSAEReservedBackwardsCompatibility', 0, 2),
+        BitField('preEstablishedKeyUsed', 0, 1),
+        BitField('encryptedMessage', 0, 1),
+        BitField('signedMessage', 0, 1),
+        BitField('signedResponseRequested', 0, 1),
+        BitField('ISOSAEReserved', 0, 9),
+        ByteField('signatureEncryptionCalculation', 0),
+        XShortField('signatureLength', 0),
+        XShortField('antiReplayCounter', 0),
+        ByteField('internalMessageServiceRequestId', 0),
+        StrField('dataRecord', b"", fmt="B")
     ]
 
 
@@ -351,7 +522,18 @@ bind_layers(UDS, UDS_SDT, service=0x84)
 class UDS_SDTPR(Packet):
     name = 'SecuredDataTransmissionPositiveResponse'
     fields_desc = [
-        StrField('securityDataResponseRecord', b"")
+        BitField('requestMessage', 0, 1),
+        BitField('ISOSAEReservedBackwardsCompatibility', 0, 2),
+        BitField('preEstablishedKeyUsed', 0, 1),
+        BitField('encryptedMessage', 0, 1),
+        BitField('signedMessage', 0, 1),
+        BitField('signedResponseRequested', 0, 1),
+        BitField('ISOSAEReserved', 0, 9),
+        ByteField('signatureEncryptionCalculation', 0),
+        XShortField('signatureLength', 0),
+        XShortField('antiReplayCounter', 0),
+        ByteField('internalMessageServiceResponseId', 0),
+        StrField('dataRecord', b"", fmt="B")
     ]
 
     def answers(self, other):
@@ -566,6 +748,7 @@ bind_layers(UDS, UDS_RSDBIPR, service=0x64)
 
 # #########################RDBPI###################################
 class UDS_RDBPI(Packet):
+    periodicDataIdentifiers = ObservableDict()
     transmissionModes = {
         0: 'ISOSAEReserved',
         1: 'sendAtSlowRate',
@@ -576,7 +759,7 @@ class UDS_RDBPI(Packet):
     name = 'ReadDataByPeriodicIdentifier'
     fields_desc = [
         ByteEnumField('transmissionMode', 0, transmissionModes),
-        ByteField('periodicDataIdentifier', 0),
+        ByteEnumField('periodicDataIdentifier', 0, periodicDataIdentifiers),
         StrField('furtherPeriodicDataIdentifier', b"", fmt="B")
     ]
 
@@ -1033,7 +1216,7 @@ class UDS_RFT(Packet):
 
     fields_desc = [
         XByteEnumField('modeOfOperation', 0, modeOfOperations),
-        FieldLenField('filePathAndNameLength', 0,
+        FieldLenField('filePathAndNameLength', None,
                       length_of='filePathAndName', fmt='H'),
         StrLenField('filePathAndName', b"",
                     length_from=lambda p: p.filePathAndNameLength),
@@ -1041,7 +1224,8 @@ class UDS_RFT(Packet):
                          lambda p: p.modeOfOperation not in [2, 5]),
         ConditionalField(BitField('encryptingMethod', 0, 4),
                          lambda p: p.modeOfOperation not in [2, 5]),
-        ConditionalField(FieldLenField('fileSizeParameterLength', 0, fmt="B",
+        ConditionalField(FieldLenField('fileSizeParameterLength', None,
+                                       fmt="B",
                                        length_of='fileSizeUnCompressed'),
                          lambda p: UDS_RFT._contains_file_size(p)),
         ConditionalField(StrLenField('fileSizeUnCompressed', b"",
@@ -1067,7 +1251,7 @@ class UDS_RFTPR(Packet):
 
     fields_desc = [
         XByteEnumField('modeOfOperation', 0, UDS_RFT.modeOfOperations),
-        ConditionalField(FieldLenField('lengthFormatIdentifier', 0,
+        ConditionalField(FieldLenField('lengthFormatIdentifier', None,
                                        length_of='maxNumberOfBlockLength',
                                        fmt='B'),
                          lambda p: p.modeOfOperation != 2),
@@ -1078,7 +1262,8 @@ class UDS_RFTPR(Packet):
                          lambda p: p.modeOfOperation != 0x02),
         ConditionalField(BitField('encryptingMethod', 0, 4),
                          lambda p: p.modeOfOperation != 0x02),
-        ConditionalField(FieldLenField('fileSizeOrDirInfoParameterLength', 0,
+        ConditionalField(FieldLenField('fileSizeOrDirInfoParameterLength',
+                         None,
                          length_of='fileSizeUncompressedOrDirInfoLength'),
                          lambda p: p.modeOfOperation not in [1, 2, 3]),
         ConditionalField(StrLenField('fileSizeUncompressedOrDirInfoLength',
@@ -1149,6 +1334,7 @@ class UDS_NR(Packet):
         0x35: 'invalidKey',
         0x36: 'exceedNumberOfAttempts',
         0x37: 'requiredTimeDelayNotExpired',
+        0x3A: 'secureDataVerificationFailed',
         0x70: 'uploadDownloadNotAccepted',
         0x71: 'transferDataSuspended',
         0x72: 'generalProgrammingFailure',
@@ -1210,72 +1396,14 @@ class UDS_TesterPresentSender(PeriodicSenderThread):
 
     def run(self):
         # type: () -> None
-        while not self._stopped.is_set():
+        while not self._stopped.is_set() and not self._socket.closed:
             for p in self._pkts:
-                self._socket.sr1(p, timeout=0.3, verbose=False)
+                try:
+                    self._socket.sr1(p, timeout=0.3, verbose=False)
+                except (OSError, ValueError, Scapy_Exception) as e:
+                    log_automotive.exception(
+                        "Exception in TesterPresentSender: %s", e)
+                    break
                 time.sleep(self._interval)
-
-
-def UDS_SessionEnumerator(sock, session_range=range(0x100), reset_wait=1.5):
-    """ Enumerates session ID's in given range
-        and returns list of UDS()/UDS_DSC() packets
-        with valid session types
-
-    Args:
-        sock: socket where packets are sent
-        session_range: range for session ID's
-        reset_wait: wait time in sec after every packet
-    """
-    pkts = (req for tup in
-            product(UDS() / UDS_DSC(diagnosticSessionType=session_range),
-                    UDS() / UDS_ER(resetType='hardReset')) for req in tup)
-    results, _ = sock.sr(pkts, timeout=len(session_range) * reset_wait * 2 + 1,
-                         verbose=False, inter=reset_wait)
-    return [req for req, res in results if req is not None and
-            req.service != 0x11 and
-            (res.service == 0x50 or
-             res.negativeResponseCode not in [0x10, 0x11, 0x12])]
-
-
-def UDS_ServiceEnumerator(sock, session="DefaultSession",
-                          filter_responses=True):
-    """ Enumerates every service ID
-        and returns list of tuples. Each tuple contains
-        the session and the respective positive response
-
-    Args:
-        sock: socket where packet is sent periodically
-        session: session in which the services are enumerated
-    """
-    pkts = (UDS(service=x) for x in set(x & ~0x40 for x in range(0x100)))
-    found_services = sock.sr(pkts, timeout=5, verbose=False)
-    return [(session, p) for _, p in found_services[0] if
-            p.service != 0x7f or
-            (p.negativeResponseCode not in [0x10, 0x11] or not
-            filter_responses)]
-
-
-def getTableEntry(tup):
-    """ Helping function for make_lined_table.
-        Returns the session and response code of tup.
-
-    Args:
-        tup: tuple with session and UDS response package
-
-    Example:
-        make_lined_table([('DefaultSession', UDS()/UDS_SAPR(),
-                           'ExtendedDiagnosticSession', UDS()/UDS_IOCBI())],
-                           getTableEntry)
-    """
-    session, pkt = tup
-    if pkt.service == 0x7f:
-        return (session,
-                "0x%02x: %s" % (pkt.requestServiceId,
-                                pkt.sprintf("%UDS_NR.requestServiceId%")),
-                pkt.sprintf("%UDS_NR.negativeResponseCode%"))
-    else:
-        return (session,
-                "0x%02x: %s" % (pkt.service & ~0x40,
-                                pkt.get_field('service').
-                                i2s[pkt.service & ~0x40]),
-                "PositiveResponse")
+                if self._stopped.is_set() or self._socket.closed:
+                    break
