@@ -1,7 +1,9 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
+# See https://scapy.net/ for more information
 # Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
 #               2015, 2016, 2017 Maxence Tury
-# This program is published under a GPLv2 license
+#               2019 Romain Perez
 
 """
 TLS handshake fields & logic.
@@ -12,16 +14,30 @@ mechanisms which are addressed with keyexchange.py.
 
 from __future__ import absolute_import
 import math
+import os
 import struct
 
 from scapy.error import log_runtime, warning
-from scapy.fields import ByteEnumField, ByteField, EnumField, Field, \
-    FieldLenField, IntField, PacketField, PacketListField, ShortField, \
-    StrFixedLenField, StrLenField, ThreeBytesField, UTCTimeField
+from scapy.fields import (
+    ByteEnumField,
+    ByteField,
+    Field,
+    FieldLenField,
+    IntField,
+    PacketField,
+    PacketLenField,
+    PacketListField,
+    ShortEnumField,
+    ShortField,
+    StrFixedLenField,
+    StrLenField,
+    ThreeBytesField,
+    UTCTimeField,
+)
 
 from scapy.compat import hex_bytes, orb, raw
 from scapy.config import conf
-from scapy.modules import six
+from scapy.libs import six
 from scapy.packet import Packet, Raw, Padding
 from scapy.utils import randstring, repr_hex
 from scapy.layers.x509 import OCSP_Response
@@ -29,16 +45,21 @@ from scapy.layers.tls.cert import Cert
 from scapy.layers.tls.basefields import (_tls_version, _TLSVersionField,
                                          _TLSClientVersionField)
 from scapy.layers.tls.extensions import (_ExtensionsLenField, _ExtensionsField,
-                                         _cert_status_type, TLS_Ext_SupportedVersion_CH,  # noqa: E501
+                                         _cert_status_type,
+                                         TLS_Ext_SupportedVersion_CH,
                                          TLS_Ext_SignatureAlgorithms,
                                          TLS_Ext_SupportedVersion_SH,
-                                         TLS_Ext_EarlyDataIndication)
+                                         TLS_Ext_EarlyDataIndication,
+                                         _tls_hello_retry_magic,
+                                         TLS_Ext_ExtendedMasterSecret,
+                                         TLS_Ext_EncryptThenMAC)
 from scapy.layers.tls.keyexchange import (_TLSSignature, _TLSServerParamsField,
                                           _TLSSignatureField, ServerRSAParams,
                                           SigAndHashAlgsField, _tls_hash_sig,
                                           SigAndHashAlgsLenField)
 from scapy.layers.tls.session import (_GenericTLSSessionInheritance,
                                       readConnState, writeConnState)
+from scapy.layers.tls.keyexchange_tls13 import TLS_Ext_PreSharedKey_CH
 from scapy.layers.tls.crypto.compression import (_tls_compression_algs,
                                                  _tls_compression_algs_cls,
                                                  Comp_NULL, _GenericComp,
@@ -47,6 +68,11 @@ from scapy.layers.tls.crypto.suites import (_tls_cipher_suites,
                                             _tls_cipher_suites_cls,
                                             _GenericCipherSuite,
                                             _GenericCipherSuiteMetaclass)
+from scapy.layers.tls.crypto.hkdf import TLS13_HKDF
+
+if conf.crypto_valid:
+    from cryptography.hazmat.backends import default_backend
+    from cryptography.hazmat.primitives import hashes
 
 
 ###############################################################################
@@ -89,6 +115,7 @@ class _TLSHandshake(_GenericTLSSessionInheritance):
         """
         Covers both post_build- and post_dissection- context updates.
         """
+
         self.tls_session.handshake_messages.append(msg_str)
         self.tls_session.handshake_messages_parsed.append(self)
 
@@ -125,6 +152,9 @@ class _GMTUnixTimeField(UTCTimeField):
             return x
         return 0
 
+    def i2m(self, pkt, x):
+        return int(x) if x is not None else 0
+
 
 class _TLSRandomBytesField(StrFixedLenField):
     def i2repr(self, pkt, x):
@@ -155,8 +185,7 @@ class _CipherSuitesField(StrLenField):
             s2i[dico[k]] = k
 
     def any2i_one(self, pkt, x):
-        if (isinstance(x, _GenericCipherSuite) or
-                isinstance(x, _GenericCipherSuiteMetaclass)):
+        if isinstance(x, (_GenericCipherSuite, _GenericCipherSuiteMetaclass)):
             x = x.val
         if isinstance(x, bytes):
             x = self.s2i[x]
@@ -205,8 +234,7 @@ class _CipherSuitesField(StrLenField):
 class _CompressionMethodsField(_CipherSuitesField):
 
     def any2i_one(self, pkt, x):
-        if (isinstance(x, _GenericComp) or
-                isinstance(x, _GenericCompMetaclass)):
+        if isinstance(x, (_GenericComp, _GenericCompMetaclass)):
             x = x.val
         if isinstance(x, str):
             x = self.s2i[x]
@@ -286,9 +314,12 @@ class TLSClientHello(_TLSHandshake):
         along with the raw string representing this handshake message.
         """
         super(TLSClientHello, self).tls_session_update(msg_str)
-
         s = self.tls_session
         s.advertised_tls_version = self.version
+        # This ClientHello could be a 1.3 one. Let's store the sid
+        # in all cases
+        if self.sidlen and self.sidlen > 0:
+            s.sid = self.sid
         self.random_bytes = msg_str[10:38]
         s.client_random = (struct.pack('!I', self.gmt_unix_time) +
                            self.random_bytes)
@@ -299,13 +330,20 @@ class TLSClientHello(_TLSHandshake):
         if self.ext:
             for e in self.ext:
                 if isinstance(e, TLS_Ext_SupportedVersion_CH):
-                    s.advertised_tls_version = e.versions[0]
+                    for ver in sorted(e.versions, reverse=True):
+                        # RFC 8701: GREASE of TLS will send unknown versions
+                        # here. We have to ignore them
+                        if ver in _tls_version:
+                            s.advertised_tls_version = ver
+                            break
+                    if s.sid:
+                        s.middlebox_compatibility = True
 
                 if isinstance(e, TLS_Ext_SignatureAlgorithms):
                     s.advertised_sig_algs = e.sig_algs
 
 
-class TLS13ClientHello(TLSClientHello):
+class TLS13ClientHello(_TLSHandshake):
     """
     TLS 1.3 ClientHello, with abilities to handle extensions.
 
@@ -345,7 +383,66 @@ class TLS13ClientHello(TLSClientHello):
     def post_build(self, p, pay):
         if self.random_bytes is None:
             p = p[:6] + randstring(32) + p[6 + 32:]
-        return super(TLS13ClientHello, self).post_build(p, pay)
+        # We don't call the post_build function from class _TLSHandshake
+        # to compute the message length because we need that value now
+        # for the HMAC in binder
+        tmp_len = len(p)
+        if self.msglen is None:
+            sz = tmp_len - 4
+            p = struct.pack("!I", (orb(p[0]) << 24) | sz) + p[4:]
+        s = self.tls_session
+        if self.ext:
+            for e in self.ext:
+                if isinstance(e, TLS_Ext_PreSharedKey_CH):
+                    if s.client_session_ticket:
+                        # For a resumed PSK, the hash function use
+                        # to compute the binder must be the same
+                        # as the one used to establish the original
+                        # conntection. For that, we assume that
+                        # the ciphersuite associate with the ticket
+                        # is given as argument to tlsSession
+                        # (see layers/tls/automaton_cli.py for an
+                        # example)
+                        res_suite = s.tls13_ticket_ciphersuite
+                        cs_cls = _tls_cipher_suites_cls[res_suite]
+                        hkdf = TLS13_HKDF(cs_cls.hash_alg.name.lower())
+                        hash_len = hkdf.hash.digest_size
+                        s.compute_tls13_early_secrets(external=False)
+                    else:
+                        # For out of band PSK, SHA-256 is used as default
+                        # hash functions for HKDF
+                        hkdf = TLS13_HKDF("sha256")
+                        hash_len = hkdf.hash.digest_size
+                        s.compute_tls13_early_secrets(external=True)
+
+                    # RFC8446 4.2.11.2
+                    # "Each entry in the binders list is computed as an HMAC
+                    # over a transcript hash (see Section 4.4.1) containing a
+                    # partial ClientHello up to and including the
+                    # PreSharedKeyExtension.identities field."
+                    # PSK Binders field is :
+                    #   - PSK Binders length (2 bytes)
+                    #   - First PSK Binder length (1 byte) +
+                    #         HMAC (hash_len bytes)
+                    # The PSK Binder is computed in the same way as the
+                    # Finished message with binder_key as BaseKey
+
+                    handshake_context = b""
+                    if s.tls13_retry:
+                        for m in s.handshake_messages:
+                            handshake_context += m
+                    handshake_context += p[:-hash_len - 3]
+
+                    binder_key = s.tls13_derived_secrets["binder_key"]
+                    psk_binder = hkdf.compute_verify_data(binder_key,
+                                                          handshake_context)
+
+                    # Here, we replaced the last 32 bytes of the packet by the
+                    # new HMAC values computed over the ClientHello (without
+                    # the binders)
+                    p = p[:-hash_len] + psk_binder
+
+        return p + pay
 
     def tls_session_update(self, msg_str):
         """
@@ -357,12 +454,19 @@ class TLS13ClientHello(TLSClientHello):
 
         if self.sidlen and self.sidlen > 0:
             s.sid = self.sid
-        self.random_bytes = msg_str[10:38]
+            s.middlebox_compatibility = True
+
+        self.random_bytes = msg_str[6:38]
         s.client_random = self.random_bytes
         if self.ext:
             for e in self.ext:
                 if isinstance(e, TLS_Ext_SupportedVersion_CH):
-                    self.tls_session.advertised_tls_version = e.versions[0]
+                    for ver in sorted(e.versions, reverse=True):
+                        # RFC 8701: GREASE of TLS will send unknown versions
+                        # here. We have to ignore them
+                        if ver in _tls_version:
+                            self.tls_session.advertised_tls_version = ver
+                            break
                 if isinstance(e, TLS_Ext_SignatureAlgorithms):
                     s.advertised_sig_algs = e.sig_algs
 
@@ -395,7 +499,7 @@ class TLSServerHello(_TLSHandshake):
                    _SessionIDField("sid", "",
                                    length_from=lambda pkt: pkt.sidlen),
 
-                   EnumField("cipher", None, _tls_cipher_suites),
+                   ShortEnumField("cipher", None, _tls_cipher_suites),
                    _CompressionMethodsField("comp", [0],
                                             _tls_compression_algs,
                                             itemfmt="B",
@@ -403,10 +507,9 @@ class TLSServerHello(_TLSHandshake):
 
                    _ExtensionsLenField("extlen", None, length_of="ext"),
                    _ExtensionsField("ext", None,
-                                    length_from=lambda pkt: (pkt.msglen -
-                                                             (pkt.sidlen or 0) -  # noqa: E501
-                                                             38))]
-    # 40)) ]
+                                    length_from=lambda pkt: (
+                                        pkt.msglen - (pkt.sidlen or 0) - 40
+                                    ))]
 
     @classmethod
     def dispatch_hook(cls, _pkt=None, *args, **kargs):
@@ -433,18 +536,28 @@ class TLSServerHello(_TLSHandshake):
         """
         super(TLSServerHello, self).tls_session_update(msg_str)
 
-        self.tls_session.tls_version = self.version
-        self.random_bytes = msg_str[10:38]
-        self.tls_session.server_random = (struct.pack('!I',
-                                                      self.gmt_unix_time) +
-                                          self.random_bytes)
-        self.tls_session.sid = self.sid
+        s = self.tls_session
+        s.tls_version = self.version
+        if hasattr(self, 'gmt_unix_time'):
+            self.random_bytes = msg_str[10:38]
+            s.server_random = (struct.pack('!I', self.gmt_unix_time) +
+                               self.random_bytes)
+        else:
+            s.server_random = self.random_bytes
+        s.sid = self.sid
+
+        if self.ext:
+            for e in self.ext:
+                if isinstance(e, TLS_Ext_ExtendedMasterSecret):
+                    self.tls_session.extms = True
+                if isinstance(e, TLS_Ext_EncryptThenMAC):
+                    self.tls_session.encrypt_then_mac = True
 
         cs_cls = None
         if self.cipher:
             cs_val = self.cipher
             if cs_val not in _tls_cipher_suites_cls:
-                warning("Unknown cipher suite %d from ServerHello" % cs_val)
+                warning("Unknown cipher suite %d from ServerHello", cs_val)
                 # we do not try to set a default nor stop the execution
             else:
                 cs_cls = _tls_cipher_suites_cls[cs_val]
@@ -453,20 +566,20 @@ class TLSServerHello(_TLSHandshake):
         if self.comp:
             comp_val = self.comp[0]
             if comp_val not in _tls_compression_algs_cls:
-                err = "Unknown compression alg %d from ServerHello" % comp_val
-                warning(err)
+                err = "Unknown compression alg %d from ServerHello"
+                warning(err, comp_val)
                 comp_val = 0
             comp_cls = _tls_compression_algs_cls[comp_val]
 
-        connection_end = self.tls_session.connection_end
-        self.tls_session.pwcs = writeConnState(ciphersuite=cs_cls,
-                                               compression_alg=comp_cls,
-                                               connection_end=connection_end,
-                                               tls_version=self.version)
-        self.tls_session.prcs = readConnState(ciphersuite=cs_cls,
-                                              compression_alg=comp_cls,
-                                              connection_end=connection_end,
-                                              tls_version=self.version)
+        connection_end = s.connection_end
+        s.pwcs = writeConnState(ciphersuite=cs_cls,
+                                compression_alg=comp_cls,
+                                connection_end=connection_end,
+                                tls_version=self.version)
+        s.prcs = readConnState(ciphersuite=cs_cls,
+                               compression_alg=comp_cls,
+                               connection_end=connection_end,
+                               tls_version=self.version)
 
 
 _tls_13_server_hello_fields = [
@@ -477,7 +590,7 @@ _tls_13_server_hello_fields = [
     FieldLenField("sidlen", None, length_of="sid", fmt="B"),
     _SessionIDField("sid", "",
                     length_from=lambda pkt: pkt.sidlen),
-    EnumField("cipher", None, _tls_cipher_suites),
+    ShortEnumField("cipher", None, _tls_cipher_suites),
     _CompressionMethodsField("comp", [0],
                              _tls_compression_algs,
                              itemfmt="B",
@@ -489,11 +602,23 @@ _tls_13_server_hello_fields = [
 ]
 
 
-class TLS13ServerHello(_TLSHandshake):
+class TLS13ServerHello(TLSServerHello):
     """ TLS 1.3 ServerHello """
     name = "TLS 1.3 Handshake - Server Hello"
     fields_desc = _tls_13_server_hello_fields
 
+    # ServerHello and HelloRetryRequest has the same structure and the same
+    # msgId. We need to check the server_random value to determine which it is.
+    @classmethod
+    def dispatch_hook(cls, _pkt=None, *args, **kargs):
+        if _pkt and len(_pkt) >= 38:
+            # If SHA-256("HelloRetryRequest") == server_random,
+            # this message is a HelloRetryRequest
+            random_bytes = _pkt[6:38]
+            if random_bytes == _tls_hello_retry_magic:
+                return TLS13HelloRetryRequest
+        return TLS13ServerHello
+
     def post_build(self, p, pay):
         if self.random_bytes is None:
             p = p[:6] + randstring(32) + p[6 + 32:]
@@ -506,38 +631,48 @@ class TLS13ServerHello(_TLSHandshake):
         cipher suite (if recognized), and finally we instantiate the write and
         read connection states.
         """
-        super(TLS13ServerHello, self).tls_session_update(msg_str)
-
         s = self.tls_session
+        s.server_random = self.random_bytes
+        s.ciphersuite = self.cipher
+        s.tls_version = self.version
+        # Check extensions
         if self.ext:
             for e in self.ext:
                 if isinstance(e, TLS_Ext_SupportedVersion_SH):
                     s.tls_version = e.version
                     break
-        s.server_random = self.random_bytes
-        s.ciphersuite = self.cipher
+
+        if s.tls_version < 0x304:
+            # This means that the server does not support TLS 1.3 and ignored
+            # the initial TLS 1.3 ClientHello. tls_version has been updated
+            return TLSServerHello.tls_session_update(self, msg_str)
+        else:
+            _TLSHandshake.tls_session_update(self, msg_str)
 
         cs_cls = None
         if self.cipher:
             cs_val = self.cipher
             if cs_val not in _tls_cipher_suites_cls:
-                warning("Unknown cipher suite %d from ServerHello" % cs_val)
+                warning("Unknown cipher suite %d from ServerHello", cs_val)
                 # we do not try to set a default nor stop the execution
             else:
                 cs_cls = _tls_cipher_suites_cls[cs_val]
 
         connection_end = s.connection_end
-
         if connection_end == "server":
             s.pwcs = writeConnState(ciphersuite=cs_cls,
                                     connection_end=connection_end,
                                     tls_version=s.tls_version)
-            s.triggered_pwcs_commit = True
+
+            if not s.middlebox_compatibility:
+                s.triggered_pwcs_commit = True
         elif connection_end == "client":
+
             s.prcs = readConnState(ciphersuite=cs_cls,
                                    connection_end=connection_end,
                                    tls_version=s.tls_version)
-            s.triggered_prcs_commit = True
+            if not s.middlebox_compatibility:
+                s.triggered_prcs_commit = True
 
         if s.tls13_early_secret is None:
             # In case the connState was not pre-initialized, we could not
@@ -561,6 +696,37 @@ class TLS13HelloRetryRequest(_TLSHandshake):
 
     fields_desc = _tls_13_server_hello_fields
 
+    def build(self):
+        fval = self.getfieldval("random_bytes")
+        if fval is None:
+            self.random_bytes = _tls_hello_retry_magic
+        return _TLSHandshake.build(self)
+
+    def tls_session_update(self, msg_str):
+        s = self.tls_session
+        s.tls13_retry = True
+        s.tls13_client_pubshares = {}
+        # If the server responds to a ClientHello with a HelloRetryRequest
+        # The value of the first ClientHello is replaced by a message_hash
+        if s.client_session_ticket:
+            cs_cls = _tls_cipher_suites_cls[s.tls13_ticket_ciphersuite]
+            hkdf = TLS13_HKDF(cs_cls.hash_alg.name.lower())
+            hash_len = hkdf.hash.digest_size
+        else:
+            cs_cls = _tls_cipher_suites_cls[self.cipher]
+            hkdf = TLS13_HKDF(cs_cls.hash_alg.name.lower())
+            hash_len = hkdf.hash.digest_size
+
+        handshake_context = struct.pack("B", 254)
+        handshake_context += struct.pack("B", 0)
+        handshake_context += struct.pack("B", 0)
+        handshake_context += struct.pack("B", hash_len)
+        digest = hashes.Hash(hkdf.hash, backend=default_backend())
+        digest.update(s.handshake_messages[0])
+        handshake_context += digest.finalize()
+        s.handshake_messages[0] = handshake_context
+        super(TLS13HelloRetryRequest, self).tls_session_update(msg_str)
+
 
 ###############################################################################
 #   EncryptedExtensions                                                       #
@@ -595,16 +761,18 @@ class TLSEncryptedExtensions(_TLSHandshake):
         # EndOfEarlyData message
         if connection_end == "server":
             if not early_data_accepted:
-                s.prcs = writeConnState(ciphersuite=type(s.wcs.ciphersuite),
-                                        connection_end=connection_end,
-                                        tls_version=s.tls_version)
+                s.prcs = readConnState(ciphersuite=type(s.wcs.ciphersuite),
+                                       connection_end=connection_end,
+                                       tls_version=s.tls_version)
 
-                s.triggered_prcs_commit = True
                 chts = s.tls13_derived_secrets["client_handshake_traffic_secret"]  # noqa: E501
                 s.prcs.tls13_derive_keys(chts)
 
-                s.rcs = self.tls_session.prcs
-                s.triggered_prcs_commit = False
+                if not s.middlebox_compatibility:
+                    s.rcs = self.tls_session.prcs
+                    s.triggered_prcs_commit = False
+                else:
+                    s.triggered_prcs_commit = True
 
     def post_dissection_tls_session_update(self, msg_str):
         self.tls_session_update(msg_str)
@@ -628,13 +796,13 @@ class TLSEncryptedExtensions(_TLSHandshake):
                 s.pwcs = writeConnState(ciphersuite=type(s.rcs.ciphersuite),
                                         connection_end=connection_end,
                                         tls_version=s.tls_version)
-
-                s.triggered_pwcs_commit = True
                 chts = s.tls13_derived_secrets["client_handshake_traffic_secret"]  # noqa: E501
                 s.pwcs.tls13_derive_keys(chts)
-
-                s.wcs = self.tls_session.pwcs
-                s.triggered_pwcs_commit = False
+                if not s.middlebox_compatibility:
+                    s.wcs = self.tls_session.pwcs
+                    s.triggered_pwcs_commit = False
+                else:
+                    s.triggered_prcs_commit = True
 ###############################################################################
 #   Certificate                                                               #
 ###############################################################################
@@ -689,9 +857,9 @@ class _ASN1CertListField(StrLenField):
         if tmp_len is not None:
             m, ret = s[:tmp_len], s[tmp_len:]
         while m:
-            clen = struct.unpack("!I", b'\x00' + m[:3])[0]
-            lst.append((clen, Cert(m[3:3 + clen])))
-            m = m[3 + clen:]
+            c_len = struct.unpack("!I", b'\x00' + m[:3])[0]
+            lst.append((c_len, Cert(m[3:3 + c_len])))
+            m = m[3 + c_len:]
         return m + ret, lst
 
     def i2m(self, pkt, i):
@@ -734,9 +902,9 @@ class _ASN1CertField(StrLenField):
         m = s
         if tmp_len is not None:
             m, ret = s[:tmp_len], s[tmp_len:]
-        clen = struct.unpack("!I", b'\x00' + m[:3])[0]
-        len_cert = (clen, Cert(m[3:3 + clen]))
-        m = m[3 + clen:]
+        c_len = struct.unpack("!I", b'\x00' + m[:3])[0]
+        len_cert = (c_len, Cert(m[3:3 + c_len]))
+        m = m[3 + c_len:]
         return m + ret, len_cert
 
     def i2m(self, pkt, i):
@@ -844,7 +1012,7 @@ class TLSServerKeyExchange(_TLSHandshake):
                                       length_from=lambda pkt: pkt.msglen - len(pkt.params))]  # noqa: E501
 
     def build(self, *args, **kargs):
-        """
+        r"""
         We overload build() method in order to provide a valid default value
         for params based on TLS session if not provided. This cannot be done by
         overriding i2m() because the method is called on a copy of the packet.
@@ -852,16 +1020,17 @@ class TLSServerKeyExchange(_TLSHandshake):
         The 'params' field is built according to key_exchange.server_kx_msg_cls
         which should have been set after receiving a cipher suite in a
         previous ServerHello. Usual cases are:
+
         - None: for RSA encryption or fixed FF/ECDH. This should never happen,
           as no ServerKeyExchange should be generated in the first place.
         - ServerDHParams: for ephemeral FFDH. In that case, the parameter to
           server_kx_msg_cls does not matter.
-        - ServerECDH*Params: for ephemeral ECDH. There are actually three
+        - ServerECDH\*Params: for ephemeral ECDH. There are actually three
           classes, which are dispatched by _tls_server_ecdh_cls_guess on
           the first byte retrieved. The default here is b"\03", which
           corresponds to ServerECDHNamedCurveParams (implicit curves).
 
-        When the Server*DHParams are built via .fill_missing(), the session
+        When the Server\*DHParams are built via .fill_missing(), the session
         server_kx_privkey will be updated accordingly.
         """
         fval = self.getfieldval("params")
@@ -878,7 +1047,6 @@ class TLSServerKeyExchange(_TLSHandshake):
                 except Exception:
                     if conf.debug_dissector:
                         raise
-                    pass
             else:
                 cls = Raw()
             self.params = cls
@@ -886,7 +1054,7 @@ class TLSServerKeyExchange(_TLSHandshake):
         fval = self.getfieldval("sig")
         if fval is None:
             s = self.tls_session
-            if s.pwcs:
+            if s.pwcs and s.client_random:
                 if not s.pwcs.key_exchange.anonymous:
                     p = self.params
                     if p is None:
@@ -992,7 +1160,7 @@ class TLSCertificateRequest(_TLSHandshake):
                    SigAndHashAlgsLenField("sig_algs_len", None,
                                           length_of="sig_algs"),
                    SigAndHashAlgsField("sig_algs", [0x0403, 0x0401, 0x0201],
-                                       EnumField("hash_sig", None, _tls_hash_sig),  # noqa: E501
+                                       ShortEnumField("hash_sig", None, _tls_hash_sig),  # noqa: E501
                                        length_from=lambda pkt: pkt.sig_algs_len),  # noqa: E501
                    FieldLenField("certauthlen", None, fmt="!H",
                                  length_of="certauth"),
@@ -1012,6 +1180,7 @@ class TLS13CertificateRequest(_TLSHandshake):
                    _ExtensionsField("ext", None,
                                     length_from=lambda pkt: pkt.msglen -
                                     pkt.cert_req_ctxt_len - 3)]
+
 ###############################################################################
 #   ServerHelloDone                                                           #
 ###############################################################################
@@ -1039,7 +1208,10 @@ class TLSCertificateVerify(_TLSHandshake):
         if sig is None:
             s = self.tls_session
             m = b"".join(s.handshake_messages)
-            if s.tls_version >= 0x0304:
+            tls_version = s.tls_version
+            if tls_version is None:
+                tls_version = s.advertised_tls_version
+            if tls_version >= 0x0304:
                 if s.connection_end == "client":
                     context_string = b"TLS 1.3, client CertificateVerify"
                 elif s.connection_end == "server":
@@ -1056,7 +1228,10 @@ class TLSCertificateVerify(_TLSHandshake):
     def post_dissection(self, pkt):
         s = self.tls_session
         m = b"".join(s.handshake_messages)
-        if s.tls_version >= 0x0304:
+        tls_version = s.tls_version
+        if tls_version is None:
+            tls_version = s.advertised_tls_version
+        if tls_version >= 0x0304:
             if s.connection_end == "client":
                 context_string = b"TLS 1.3, server CertificateVerify"
             elif s.connection_end == "server":
@@ -1086,9 +1261,9 @@ class _TLSCKExchKeysField(PacketField):
     __slots__ = ["length_from"]
     holds_packet = 1
 
-    def __init__(self, name, length_from=None, remain=0):
+    def __init__(self, name, length_from=None):
         self.length_from = length_from
-        PacketField.__init__(self, name, None, None, remain=remain)
+        PacketField.__init__(self, name, None, None)
 
     def m2i(self, pkt, m):
         """
@@ -1134,6 +1309,30 @@ class TLSClientKeyExchange(_TLSHandshake):
             self.exchkeys = cls
         return _TLSHandshake.build(self, *args, **kargs)
 
+    def tls_session_update(self, msg_str):
+        """
+        Finalize the EXTMS messages and compute the hash
+        """
+        super(TLSClientKeyExchange, self).tls_session_update(msg_str)
+
+        if self.tls_session.extms:
+            to_hash = b''.join(self.tls_session.handshake_messages)
+            # https://tools.ietf.org/html/rfc7627#section-3
+            if self.tls_session.tls_version >= 0x303:
+                # TLS 1.2 uses the same Hash as the PRF
+                from scapy.layers.tls.crypto.hash import _tls_hash_algs
+                hash_object = _tls_hash_algs.get(
+                    self.tls_session.prcs.prf.hash_name
+                )()
+                self.tls_session.session_hash = hash_object.digest(to_hash)
+            else:
+                # Previous TLS version use concatenation of MD5 & SHA1
+                from scapy.layers.tls.crypto.hash import Hash_MD5, Hash_SHA
+                self.tls_session.session_hash = (
+                    Hash_MD5().digest(to_hash) + Hash_SHA().digest(to_hash)
+                )
+            self.tls_session.compute_ms_and_derive_keys()
+
 
 ###############################################################################
 #   Finished                                                                  #
@@ -1162,7 +1361,10 @@ class TLSFinished(_TLSHandshake):
             s = self.tls_session
             handshake_msg = b"".join(s.handshake_messages)
             con_end = s.connection_end
-            if s.tls_version < 0x0304:
+            tls_version = s.tls_version
+            if tls_version is None:
+                tls_version = s.advertised_tls_version
+            if tls_version < 0x0304:
                 ms = s.master_secret
                 self.vdata = s.wcs.prf.compute_verify_data(con_end, "write",
                                                            handshake_msg, ms)
@@ -1174,7 +1376,10 @@ class TLSFinished(_TLSHandshake):
         s = self.tls_session
         if not s.frozen:
             handshake_msg = b"".join(s.handshake_messages)
-            if s.tls_version < 0x0304 and s.master_secret is not None:
+            tls_version = s.tls_version
+            if tls_version is None:
+                tls_version = s.advertised_tls_version
+            if tls_version < 0x0304 and s.master_secret is not None:
                 ms = s.master_secret
                 con_end = s.connection_end
                 verify_data = s.rcs.prf.compute_verify_data(con_end, "read",
@@ -1182,7 +1387,7 @@ class TLSFinished(_TLSHandshake):
                 if self.vdata != verify_data:
                     pkt_info = pkt.firstlayer().summary()
                     log_runtime.info("TLS: invalid Finished received [%s]", pkt_info)  # noqa: E501
-            elif s.tls_version >= 0x0304:
+            elif tls_version >= 0x0304:
                 con_end = s.connection_end
                 verify_data = s.compute_tls13_verify_data(con_end, "read")
                 if self.vdata != verify_data:
@@ -1192,7 +1397,10 @@ class TLSFinished(_TLSHandshake):
     def post_build_tls_session_update(self, msg_str):
         self.tls_session_update(msg_str)
         s = self.tls_session
-        if s.tls_version >= 0x0304:
+        tls_version = s.tls_version
+        if tls_version is None:
+            tls_version = s.advertised_tls_version
+        if tls_version >= 0x0304:
             s.pwcs = writeConnState(ciphersuite=type(s.wcs.ciphersuite),
                                     connection_end=s.connection_end,
                                     tls_version=s.tls_version)
@@ -1206,7 +1414,10 @@ class TLSFinished(_TLSHandshake):
     def post_dissection_tls_session_update(self, msg_str):
         self.tls_session_update(msg_str)
         s = self.tls_session
-        if s.tls_version >= 0x0304:
+        tls_version = s.tls_version
+        if tls_version is None:
+            tls_version = s.advertised_tls_version
+        if tls_version >= 0x0304:
             s.prcs = readConnState(ciphersuite=type(s.rcs.ciphersuite),
                                    connection_end=s.connection_end,
                                    tls_version=s.tls_version)
@@ -1297,7 +1508,7 @@ class ThreeBytesLenField(FieldLenField):
 _cert_status_cls = {1: OCSP_Response}
 
 
-class _StatusField(PacketField):
+class _StatusField(PacketLenField):
     def m2i(self, pkt, m):
         idtype = pkt.status_type
         cls = self.cls
@@ -1313,7 +1524,8 @@ class TLSCertificateStatus(_TLSHandshake):
                    ByteEnumField("status_type", 1, _cert_status_type),
                    ThreeBytesLenField("responselen", None,
                                       length_of="response"),
-                   _StatusField("response", None, Raw)]
+                   _StatusField("response", None, Raw,
+                                length_from=lambda pkt: pkt.responselen)]
 
 
 ###############################################################################
@@ -1416,6 +1628,31 @@ class TLS13NewSessionTicket(_TLSHandshake):
                                                              (pkt.ticketlen or 0) -  # noqa: E501
                                                              pkt.noncelen or 0) - 13)]  # noqa: E501
 
+    def build(self):
+        fval = self.getfieldval("ticket")
+        if fval == b"":
+            # Here, the ticket is just a random 48-byte label
+            # The ticket may also be a self-encrypted and self-authenticated
+            # value
+            self.ticket = os.urandom(48)
+
+        fval = self.getfieldval("ticket_nonce")
+        if fval == b"":
+            # Nonce is randomly chosen
+            self.ticket_nonce = os.urandom(32)
+
+        fval = self.getfieldval("ticket_lifetime")
+        if fval == 0xffffffff:
+            # ticket_lifetime is set to 12 hours
+            self.ticket_lifetime = 43200
+
+        fval = self.getfieldval("ticket_age_add")
+        if fval == 0:
+            # ticket_age_add is a random 32-bit value
+            self.ticket_age_add = struct.unpack("!I", os.urandom(4))[0]
+
+        return _TLSHandshake.build(self)
+
     def post_dissection_tls_session_update(self, msg_str):
         self.tls_session_update(msg_str)
         if self.tls_session.connection_end == "client":
@@ -1445,6 +1682,25 @@ class TLS13KeyUpdate(_TLSHandshake):
                    ThreeBytesField("msglen", None),
                    ByteEnumField("request_update", 0, _key_update_request)]
 
+    def post_build_tls_session_update(self, msg_str):
+        s = self.tls_session
+        s.pwcs = writeConnState(ciphersuite=type(s.wcs.ciphersuite),
+                                connection_end=s.connection_end,
+                                tls_version=s.tls_version)
+        s.triggered_pwcs_commit = True
+        s.compute_tls13_next_traffic_secrets(s.connection_end, "write")
+
+    def post_dissection_tls_session_update(self, msg_str):
+        s = self.tls_session
+        s.prcs = writeConnState(ciphersuite=type(s.rcs.ciphersuite),
+                                connection_end=s.connection_end,
+                                tls_version=s.tls_version)
+        s.triggered_prcs_commit = True
+        if s.connection_end == "server":
+            s.compute_tls13_next_traffic_secrets("client", "read")
+        elif s.connection_end == "client":
+            s.compute_tls13_next_traffic_secrets("server", "read")
+
 
 ###############################################################################
 #   All handshake messages defined in this module                             #
